-- control.lua

-- Dec 30, 2021 Mod by JustGoFly to merge OARC and Brave New World 1.1 version
-- Nov 22, 2024 Rewrite to merge with Oarc 2.x that has been modified to support bno

--local console = {
--    name = 'Console',
--    admin = true,
----    print = function(...) rcon.print(...) end,
--    print = function (...)
--      for i,v in ipairs(arg) do
--        printResult = printResult .. tostring(v) .. "\t"
--      end
--      printResult = printResult .. "\n"
--      rcon.print(printResult, {sound=defines.print_sound.never})
--    end,
--    color = {1, 1, 1, 1}
--}

require ("lib/gui_popup")
require ("lib/bno_separate_spawns")

local console = {
    name = 'Console',
    admin = true,
    print = function(...) rcon.print(...) end,
    color = {0, 1, 1, 1}    -- cyan
}

-- Main Configuration File
require("config")
require("lib/oarc_utilsX")
require("lib/bno_utils")
require("lib/init")
require("lib/auto_decon_miners")
require("lib/stats_gui")
require("lib/settings_gui")
require("lib/swarm")
require("lib/bno_config")
require("lib/bno_commands")

-- add loaders into game as default and available at start
--2.0 local function update_loaders()
--2.0     for _, force in pairs(game.forces) do
--2.0         force.recipes["loader"].enabled = force.technologies["logistics"].researched
--2.0         force.recipes["fast-loader"].enabled = force.technologies["logistics-2"].researched
--2.0         force.recipes["express-loader"].enabled = force.technologies["logistics-3"].researched
--2.0     end
--2.0 end

-- BNO player permission group to remove mining ability
local function init_permissions()
    if storage.bnoPlayerGroup==nil then
        storage.bnoPlayerGroup = game.permissions.get_group("BNO_PLAYER") or game.permissions.create_group("BNO_PLAYER") -- LuaPermissionGroup
        storage.bnoPlayerGroup.set_allows_action( defines.input_action.begin_mining,false)
        storage.bnoPlayerGroup.set_allows_action( defines.input_action.begin_mining_terrain,false)
--        storage.bnoPlayerGroup.set_allows_action( defines.input_action.craft,false)        - this won't allow selecting item from menu for putting on ground - alternative requires cheat_mode
    end
end

-- instead of using force based player.force.manual_mining_speed_modifier to remove mining - use permissions
local function configure_permissions(player)
    local charPlayer = storage.players[player.index].characterMode
    if charPlayer==nil then charPlayer = true end
    if not storage.bnoPlayerGroup then
        init_permissions()
    end
    local found=false
    if charPlayer then
        for idx, p in pairs(storage.bnoPlayerGroup.players) do   -- find this player in a permission group
            if p==player then
                storage.bnoPlayerGroup.remove_player(p)
                break
            end
        end
    else
        storage.bnoPlayerGroup.add_player(player)
    end
end

--------------------------------------------------------------------------------
-- ALL EVENT HANLDERS ARE HERE IN ONE PLACE!
--------------------------------------------------------------------------------
DISTANCE_BETWEEN_BUDDY_SPAWN = 8   -- in chunks
local oarc_scenario_interface =
{
    get_scenario_settings = function()
        storage.ocfg = remote.call("oarc_mod", "get_mod_settings")
        if (storage.ocfg) then
            log("oarc_mod settings loaded ")
            storage.ocfg.server_info.welcome_msg_title = "Brave New Oarc"
            storage.ocfg.server_info.welcome_msg = "Welcome to BNO, an integration of Oarc multiplayer and Brave New World mods with a special brew of new features."
            storage.ocfg.server_info.custom_info_header = "Brave New Oarc"
            storage.ocfg.server_info.custom_info_content = "Current game play is limited to Nauvis, and character only. More changes are planned to bring back the Brave New Players"
            storage.ocfg.server_info.discord_invite="https://discord.gg/QtkUX9kvgJ"
            
            storage.ocfg.gameplay.enable_main_team=false
            storage.ocfg.gameplay.minimum_distance_to_existing_chunks = 20
            storage.ocfg.gameplay.near_spawn_distance = 200
            storage.ocfg.gameplay.far_spawn_distance = 350
            storage.ocfg.gameplay.buddy_spawn_distance_offset = CHUNK_SIZE * DISTANCE_BETWEEN_BUDDY_SPAWN
            storage.ocfg.gameplay.minimum_distance_to_existing_chunks = DISTANCE_BETWEEN_BUDDY_SPAWN
            storage.ocfg.gameplay.enable_friendly_teams = false
--            storage.ocfg.gameplay.disable_solo_spawns = true    -- buddy Spawn only !
            storage.ocfg.surfaces_blacklist = {"Surface 1", "Surface 2", HOLDING_PEN_SURFACE_NAME}

            storage.ocfg.spawn_general.spawn_radius_tiles = CHUNK_SIZE*2.5
            if (script.active_mods["Krastorio2"]) then 
                storage.ocfg.spawn_general.spawn_radius_tiles = storage.ocfg.spawn_general.spawn_radius_tiles + 32 
            end

            if script.active_mods["VoidBlock"] or script.active_mods["stranded_block_2"] then
                storage.ocfg.resource_placement.enabled = false
            else
                storage.ocfg.resource_placement.enabled = true
            end
            storage.ocfg.resource_placement.random_order = false
            --storage.ocfg.resource_placement.vertical_offset = 48
            --storage.ocfg.resource_placement.horizontal_offset = 64
            storage.ocfg.resource_placement.linear_spacing = 8

--   if storage.players[player.index].characterMode then
            storage.ocfg.surfaces_config["nauvis"].starting_items = PLAYER_SPAWN_START_ITEMS
--   end
            storage.ocfg.surfaces_config["nauvis"].spawn_config = BNO_SPAWN_CONFIG

            -- modify biter spawns based on slider in map generator
            local starting_area = game.surfaces["nauvis"].map_gen_settings.starting_area
            if (starting_area<.5) then
                storage.ocfg.surfaces_config["nauvis"].spawn_config.safe_area.safe_radius = storage.ocfg.surfaces_config["nauvis"].spawn_config.safe_area.safe_radius * 0.5
            else
                storage.ocfg.surfaces_config["nauvis"].spawn_config.safe_area.safe_radius = storage.ocfg.surfaces_config["nauvis"].spawn_config.safe_area.safe_radius * starting_area
            end
            storage.ocfg.surfaces_config["nauvis"].spawn_config.safe_area.warn_radius = storage.ocfg.surfaces_config["nauvis"].spawn_config.safe_area.warn_radius * starting_area
            storage.ocfg.surfaces_config["nauvis"].spawn_config.safe_area.danger_radius = storage.ocfg.surfaces_config["nauvis"].spawn_config.safe_area.danger_radius * starting_area

            --=======  V O I D    B L O C K ==============--
            if script.active_mods["VoidBlock"] then

                local VoidBlock_start_items = {
                    ["coal"]                    = 10,
                    ["stone-furnace"]           = 10,
                    ["landfill"]                = 200,
                    ["small-electric-pole"]     = 10,
                    ["boiler"]                  = 1,
		            ["steam-engine"]            = 2,
		            ["gun-turret"]              = 2,
                    ["s6x-voidworks-basic"]     = 10
                }
                local VoidBlock_no_char_items = {
                    ["burner-inserter"]         = 2,
                    ["inserter"]                = 40,
                    ["transport-belt"]          = 100,
                    ["assembling-machine-1"]    = 4,
                    ["underground-belt"]        = 20,
                    ["splitter"]                = 10,
                    ["small-electric-pole"]     = 10,
                    ["landfill"]                = 600,
                    ["fast-inserter"]           = 1,
                    ["small-electric-pole"]     = 10
                }
                storage.ocfg.surfaces_config["nauvis"].starting_items.player_start_items = VoidBlock_start_items
                storage.ocfg.surfaces_config["nauvis"].starting_items.player_start_items_non_character = VoidBlock_no_char_items
                storage.ocfg.surfaces_config["nauvis"].spawn_config.solid_resources= {}
                storage.ocfg.surfaces_config["nauvis"].spawn_config.fluid_resources= {}
            --=======  S T R A N D E D  ==============--    
            elseif script.active_mods["stranded_block_2"] then
                storage.ocfg.surfaces_config["nauvis"].spawn_config.random_entities = {}
                storage.ocfg.surfaces_config["nauvis"].spawn_config.radius_modifier = 2
                storage.ocfg.surfaces_config["nauvis"].spawn_config.fill_tile = "water"
                storage.ocfg.surfaces_config["nauvis"].spawn_config.tree_entity = nil
                storage.ocfg.spawn_general.spawn_radius_tiles = 64
                local stranded_block_2_start_items = {
		            ["landfill"]                = 600,
		            ["offshore-pump"]           = 1,
		            ["chemical-plant"]          = 4,
		            ["pipe"]                    = 20,
		            ["coal"]                    = 50,
		            ["solar-panel"]             = 1,
		            ["small-electric-pole"]     = 8,
		            ["boiler"]                  = 1,
		            ["steam-engine"]            = 2,
		            ["gun-turret"]              = 2
                }
                local stranded_block_2_no_char_items = {
		            ["landfill"]                = 200,
                    ["burner-inserter"]         = 2,
                    ["inserter"]                = 20,
                    ["transport-belt"]          = 100,
                    ["assembling-machine-1"]    = 4,
                    ["underground-belt"]        = 20,
                    ["splitter"]                = 10,
                    ["small-electric-pole"]     = 10
                }

                storage.ocfg.surfaces_config["nauvis"].starting_items.player_start_items = stranded_block_2_start_items
                storage.ocfg.surfaces_config["nauvis"].starting_items.player_start_items_non_character = stranded_block_2_no_char_items

                storage.ocfg.surfaces_config["nauvis"].spawn_config.solid_resources= {}
                storage.ocfg.surfaces_config["nauvis"].spawn_config.fluid_resources= {}
            else --=======  E v e r y t h i n g   E l s e ==============--
                -- update richness of ore placed by OARC
                local spawn_config = storage.ocfg.surfaces_config["nauvis"].spawn_config
                local oarc_solid_resources = spawn_config.solid_resources

                if oarc_solid_resources == nil then
                    log("Unable to access resources to set richness from map gen slider")
                else
	                for r_name, _ in pairs(oarc_solid_resources) do -- increase the resources by the richness slider setting 
		                if (r_name and r_name ~= "") then
			                resource = game.surfaces["nauvis"].map_gen_settings.autoplace_controls[r_name]
			                if resource ~= nil then -- copy my solid resource config over to oarc - add the richness to the in base ore, based on slide
				                oarc_solid_resources[r_name].amount = BNO_SPAWN_CONFIG.solid_resources[r_name].amount * resource.richness
			                end
		                end
	                end
                end
            end
        else
            log("oarc_mod settings failed")
            storage.ocfg = {}
        end
        return storage.ocfg     -- available to Oarc to obtain my modified ocfg        
    end
}
remote.add_interface("oarc_scenario", oarc_scenario_interface)

----------------------------------------
-- On Init - only runs once the first
--   time the game starts
----------------------------------------
script.on_init(function(event)

    storage.bno = {}
	storage.spawn = {}
    storage.players = {}

-- Oarc inclusion
    if not script.active_mods["oarc-mod"] then
        error("OARC mod not found! This scenario is intended to be run with the OARC mod!")
    end

-- Init BNW
    InitBNOConfig()
    
    BNOSwarmGroupInit()
--
--    update_loaders()   -- loaders
    -- permissions to disallow mining
    init_permissions()
end)


----------------------------------------
---@class OarcModOnSpawnCreatedEvent: OarcCustomEventBase
---@field spawn_data OarcUniqueSpawn
script.on_event("oarc-mod-on-spawn-created", function(event)
    log("EVENT - oarc-mod-on-spawn-created:" .. serpent.block(event --[[@as OarcModOnSpawnCreatedEvent]]))
    -- Display starting point text as a display of dominance.
    local surface = game.surfaces[HOLDING_PEN_SURFACE_NAME]
    if (surface) then
        RenderPermanentGroundText(surface, {x=0,y=-25}, 12,    "Brave New OARC", {0.9, 0.7, 0.3, 0.8}, "center")
        if storage.bno.space_block then -- Space block
            RenderPermanentGroundText(surface, {x=0,y=5}, 4, "Space Block", {0.9, 0.7, 0.3, 0.8}, "center")  
        elseif storage.bno.seablock then -- Sea Block
            RenderPermanentGroundText(surface, {x=0,y=5}, 4, "Sea Block", {0.9, 0.7, 0.3, 0.8}, "center")
        elseif storage.bno.krastorio2 then -- Krastorio2
            RenderPermanentGroundText(surface, {x=0,y=5}, 4, "Krastorio2", {0.9, 0.7, 0.3, 0.8}, "center") 
        elseif storage.bno.alien_module then -- Alien ore
            RenderPermanentGroundText(surface, {x=0,y=5}, 4, "Alien Ore", {0.9, 0.7, 0.3, 0.8}, "center") 
        elseif storage.bno.dangOreus then -- Danger Ores
            RenderPermanentGroundText(surface, {x=0,y=5}, 4, "DangOreus", {0.9, 0.7, 0.3, 0.8}, "center") 
        elseif storage.bno.claustorephobic then -- ClaustOrePhobic
            RenderPermanentGroundText(surface, {x=0,y=5}, 4, "ClaustOrePhobic", {0.9, 0.7, 0.3, 0.8}, "center") 
        elseif storage.bno.VoidBlock then -- VoidBlock
            RenderPermanentGroundText(surface, {x=0,y=5}, 4, "VoidBlock", {0.9, 0.7, 0.3, 0.8}, "center") 
        elseif storage.bno.stranded_block_2 then -- stranded_block_2
            RenderPermanentGroundText(surface, {x=0,y=5}, 4, "Stranded Block 2", {0.9, 0.7, 0.3, 0.8}, "center") 
        end
    end
end)

----------------------------------------
script.on_load(function()
end)


----------------------------------------
-- Rocket launch event
-- Used for end game win conditions / unlocking late game stuff
----------------------------------------
script.on_event(defines.events.on_rocket_launched, function(event)
--    RocketLaunchEvent(event)
    local rocketSilo = event.rocket_silo
    player = rocketSilo.last_user
    if player.force.rockets_launched == 0 then
        firstRocketLaunched(player, event.tick)
    end
end)

----------------------------------------
-- Player may have landed on planet
-- Event variables:
--    player_index
--    old_surface_name
--    new_surface_name
----------------------------------------
script.on_event("oarc-mod-character-surface-changed", function(event)
    local player = game.players[event.player_index]
    log("Player changed surfaces: " .. player.name .. " from surface: " .. event.old_surface_name .. " to surface: " .. event.new_surface_name)
    
    -- if a BNO player (not characterMode) and he is in character mode - change to non character mode
	if ((not storage.players[player.index].characterMode) and player.character) then
          -- also we must be teleporting from platform
--     if (string.match(event.old_surface_name, "^platform")) then
--          player.character.destroy()
--          -- also teleport to the silo they left from
--          player.teleport(storage.players[player.index].return_to_pos, game.surfaces[event.new_surface_name] )
--      end
    end
end)

----------------------------------------
-- Capsule used by player resulting in an action 
----------------------------------------
script.on_event("on_trigger_created_entity", function (event)
    log("Entity created: " .. event.entity.name)
end)

----------------------------------------
-- Rocket silo mods to enable BNO player to enter rocket
----------------------------------------
local FRAME_NAME   = "space_travel_frame"
local BUTTON_NAME  = "space_travel_button"
local TARGET_SURFACE = "platform-1"   -- name of your space platform surface
local TARGET_POS     = { x = 0, y = 0 }   -- where on that surface the player arrives

----------------------------------------
-- On Gui Closed
-- For capturing player escaping custom GUI so we can close it using ESC key.
----------------------------------------
script.on_event(defines.events.on_gui_closed, function(event)
end)

local permitted_types_to_defs = {["rocket-silo"] = defines.relative_gui_type.rocket_silo_gui}

script.on_event(defines.events.on_gui_opened, function(event)
    player = game.players[event.player_index]
    if event.entity then 
        st="Event: on_gui_opened-> Player: " .. player.name
        if event.gui_type == defines.gui_type.item then
            st = st .. " item: " .. event.item.name
        elseif event.gui_type == defines.gui_type.equipment then
            st = st .. ", equipment: " .. event.equipment.name
        end
        st = st .. ", Entity: " .. event.entity.name .. ", Force: " .. event.entity.force.name
        if event.entity.last_user then
            st = st .. ", Last_User : " .. event.entity.last_user.name .. " " 
        end
        st = st .. event.entity.gps_tag
        if event.other_player then
            st = st .. ", Accessing Other player: " .. event.other_player.name
        end
        if event.element then
            st = st .. ", Accessing element: " .. event.element.name
        end
        if player.force.name ~= event.entity.force.name then
            log("ALARM: Player accessing another players entity!!!")
            log(st)
        end
    end
-- DISABLING THIS UNTIL I CAN PLAY THROUGH WITH SPACE AGE ENABLED
--    -- show if this is an entity and this player is a BNO player
--    if  event.gui_type == defines.gui_type.entity 
--    and not storage.players[event.player_index].characterMode then 
--        create_anchored_silo_gui(player, event.entity)
--    end
end)

----------------------------------------
-- Helper function to make the gui on the side of the silo 

function create_anchored_silo_gui(player, entity)
    if entity and entity.valid and entity.name == "rocket-silo" then 
        if not permitted_types_to_defs[entity.type] then return end
        local relative = player.gui.relative
        -- Avoid duplicating - delete the previously make one since content is dynamic
        if relative[FRAME_NAME] then 
            relative[FRAME_NAME].destroy()
        end
        anchor_silo_gui(player, relative, entity)
    end
end

----------------------------------------
-- Gui for rocket silo
-- Shows the travel to platform button if the platform is 
--     1) unlocked
--     2) Player has BNO starter pack
----------------------------------------
function anchor_silo_gui(player, relative, entity)
	local entity_type = entity.type
	local gui_type = permitted_types_to_defs[entity_type]

    -- build our new side-panel
	local frame = relative.add{
		name = FRAME_NAME,
		type = "frame",
		caption = {"BNO-Space-Travel"},
		direction = "vertical",
        style = "frame",
		anchor = {
			gui = gui_type,
			position = defines.relative_gui_position.right,
		},
	}

    frame.style.horizontally_stretchable = true
    frame.style.horizontal_align = "center"
    local content = frame.add{
        type = "frame",
        name = "deep_content",
        direction = "vertical",
        style = "inside_shallow_frame_with_padding"
    }        
    content.style.padding = 15
    content.style.margin = 0
    content.style.bottom_margin = 0
    show = player.force.is_space_platforms_unlocked()
    if show then
        -- [img=entity/character-corpse]    space-rocket
        content.add{
            type    = "button",
            name    = BUTTON_NAME,
            caption = "[img=item/thruster] Travel to space platform"
        }
    else
    	content.add{
            type = "label",
            caption = "Unlock your space platform by making a [img=item/space-platform-starter-pack] Space Platform Starter kit.",
	    }
    	content.add{
            type = "label",
            caption = "To bring items with you to start a new planet make a BNO Planet Starter Kit",
	    }
    end

end

----------------------------------------
-- Gui Click
----------------------------------------
script.on_event(defines.events.on_gui_click, function(event)
--    log(game.players[event.player_index].name .. " generated a gui click event: " .. tostring(evelement.name))
    -- Don't interfere with other mod related stuff.
    local player = game.players[event.player_index]

    if (event.element.get_mod() == "brave-new-oarc") then 
        if (event.element.name == "character_mode_option_checkbox") then
            storage.players[event.player_index].characterMode = event.element.state
        end
    elseif (event.element.get_mod() ~= "oarc-mod") then 
        return
    end
    if event.element==nil or not event.element.valid then return end
    if (event.element.name == "oarc_mod_gui_button") then 
        if not storage.custom_tabs_table[event.player_index].drop_index then 
            storage.custom_tabs_table[event.player_index].drop_index = 1 
        end
        -- only rebuild if we're looking at our tab
        if storage.custom_tabs_table[event.player_index].selected_tab_index == BNO_STATS_TAB_NAME then
            buildStatsTable(game.players[event.player_index], storage.custom_tabs_table[event.player_index].drop_index)
        end
        if storage.custom_tabs_table[event.player_index].selected_tab_index == BNO_SETTINGS_TAB_NAME then
            buildSettingsTable(game.players[event.player_index])
        end
        return
    elseif (event.element.name == "share_chart_checkbox") then
        if (storage.bno.share_chart[event.player_index] == nil) then
            storage.bno.share_chart[event.player_index] = true
        end
        storage.bno.share_chart[event.player_index] = not storage.bno.share_chart[event.player_index]
        local onOff = " OFF"
        if storage.bno.share_chart[event.player_index] then 
            onOff = " ON" 
        end
        player.force.share_chart=storage.bno.share_chart[event.player_index]
        player.print(player.name .. " changed the 'Sharing of Chart' option to " .. onOff)
    elseif (event.element.name == BUTTON_NAME) then 
        local platforms = {}
        for name,surface in pairs(game.surfaces) do
            if (surface.platform) then
                platforms[#platforms + 1] = surface.platform
    	        log("surface: " .. name .. " is a platform named: " .. surface.platform.name)
                if  ((surface.platform.speed== 0)
                and  (surface.platform.space_location.name == player.surface.name)) then
                    log("Player: " .. player.name .. " can go to the platform")
                    if player.character == nil then
                        storage.players[player.index].return_to_pos = player.position
                        player.create_character()   -- we must have a character to enter the silo 
                    end
                    if game.players[1].enter_space_platform(platforms[1]) then
                        log("Player: " .. player.name .." entered the space platform")
                    else
                        log("Player: " .. player.name .." FAILED to enter the space platform")
                    end
                end
            end
        end
    elseif (event.element.name == "popup_close_gui") then       -- create_popup_gui(player, "Button pressed", {"Ya man you did it", "Way to go"})
        player.gui.screen.popup_gui.destroy()
    end
end)

--=================================================================================================
script.on_event(defines.events.on_gui_selection_state_changed, function (event)
    if (event.element.name == "stats_dropdown") then
        storage.custom_tabs_table[event.player_index].drop_index = event.element.selected_index
        buildStatsTable(game.players[event.player_index], event.element.selected_index)
    end
end)


--=================================================================================================
script.on_event(defines.events.on_gui_selected_tab_changed, function(event)
    local player = game.players[event.player_index]
    local entry = storage.custom_tabs_table[event.player_index]
    -- Check if your custom tab was NOT selected
    local tab_name = event.element.tabs[event.element.selected_tab_index].tab.name
    if (tab_name == BNO_STATS_TAB_NAME) then
        if not storage.custom_tabs_table[event.player_index].drop_index then 
            storage.custom_tabs_table[event.player_index].drop_index = 1 
        end
        buildStatsTable(player, storage.custom_tabs_table[event.player_index].drop_index)
    elseif (tab_name == BNO_SETTINGS_TAB_NAME) then
        buildSettingsTable(player)
    end
    storage.custom_tabs_table[event.player_index].selected_tab_index =  tab_name   -- used in on-gui-click
end)
--=================================================================================================
script.on_event("oarc-mod-on-spawn-choices-gui-displayed", function(event)
    local player = game.players[event.player_index]
--    local sGui = player.gui
    local sGui = event.gui_element

-- Use this for character mode
    if (sGui and sGui.spawn_settings_frame ~= nil) then
        if storage.players==nil then
            storage.players[event.player_index] = {
                crafted = {},
                inventory_items = {},   
                previous_position = player.position,
                drawOnExit = nil,
                characterMode = true,
                inSpawn = true
            }
        end
        -- void snatch mod puts items in a players inventory - so no support for non character mode
            local spawn_settings_frame = sGui.spawn_settings_frame
            spawn_settings_frame.add{name = "character_mode_option_checkbox",
                    type = "checkbox",
                    caption={"bno-mode-option"},
                    state=storage.players[event.player_index].characterMode}
    end
end)

--=================================================================================================
script.on_event("oarc-mod-on-mod-top-left-gui-created", function(event)
    local player = game.players[event.player_index]
    local sGui = player.gui

    if (sGui ~= nil) then
        -- Save a reference to that tab content for later use, if it becomes nil you can ask for it again.
        -- Or you can just always ask for it when you need it, depends on how often you plan to use it.
        if storage.custom_tabs_table == nil then
            storage.custom_tabs_table = {}
        end
        if (storage.custom_tabs_table[event.player_index] == nil) then
            storage.custom_tabs_table[event.player_index] = {}
        end
        if (storage.custom_tabs_table[event.player_index].drop_index == nil) then
            storage.custom_tabs_table[event.player_index].drop_index = 1
        end

        -- Create a custom empty tab:
        remote.call("oarc_mod", "create_custom_gui_tab", player, BNO_STATS_TAB_NAME)

        -- Get the tab content element
        tab_content = remote.call("oarc_mod", "get_custom_gui_tab_content_element", player, BNO_STATS_TAB_NAME)
        storage.custom_tabs_table[event.player_index].stats_content = tab_content
        tab_content.clear()

        remote.call("oarc_mod", "create_custom_gui_tab", player, BNO_SETTINGS_TAB_NAME)
        tab_content = remote.call("oarc_mod", "get_custom_gui_tab_content_element", player, BNO_SETTINGS_TAB_NAME)
        storage.custom_tabs_table[event.player_index].bno_settings_content = tab_content
        tab_content.clear()
    end

end)


----------------------------------------
-- Player Events
----------------------------------------
script.on_event(defines.events.on_player_joined_game, function(event)
    local joiningPlayer=game.players[event.player_index]
--    if storage.bno.krastorio2 and not storage.bno.creep_initialized then
--        remote.call("kr-creep", "set_creep_on_surface", game.surfaces["oarc"].index, true)      -- oarc
--        remote.call("kr-creep", "set_creep_on_surface", game.surfaces["nauvis"].index, false)     -- nauvis
--        storage.bno.creep_initialized=true
--
--    end
--    if storage.bno.brave_new_assembling_machines and not storage.bno.surface_index_set then
--        remote.call("set_surface", "set_surface_asms_reside", game.surfaces["oarc"].index)
--        storage.bno.surface_index_set=true
--    end

    if storage.bno.dangOreus and not storage.bno.dangOreus_initialized then
-- we now work on nauvis - previously we were on oarc surface. So don't disable nauvis.
--        remote.call("dangOreus", "toggle", game.surfaces["nauvis"], false)      -- turn off ore on nauvis to save cpu
        storage.bno.dangOreus_initialized=true
    end

    -- stranded_block_2 mod hard code items into a box
    log("on_event::On Player Joined Game " .. joiningPlayer.name)
    PlayerJoinedMessages(event)
    ServerWriteFile("player_events", joiningPlayer.name .. " joined the game." .. "\n")
    -- Remove player name from map while they are online_time
    -- Render some welcoming text...

    if (storage.players[event.player_index].drawOnExit ~=nil) then
        log("Destroying drawOnExit: " .. tostring(storage.players[event.player_index].drawOnExit) .. " for player: " .. game.players[event.player_index].name)
        local RenderObj = rendering.get_object_by_id(storage.players[event.player_index].drawOnExit.id)
        if RenderObj then
            RenderObj.destroy()
        end
        storage.players[event.player_index].drawOnExit=nil
    end
    if (storage.spawn[event.player_index] == nil) then
        storage.spawn[event.player_index]={x=0,y=0}
    else
        if ((storage.spawn[event.player_index].x ~= 0) and (storage.spawn[event.player_index].y ~= 0)) then
            DisplayWelcomeBackGroundTextAtSpawn (joiningPlayer, storage.spawn[event.player_index])
        end
        if (storage.players[event.player_index].emptyInventory)  then    -- this has to be done here since Space Block on_created
            empty_players_inventory(joiningPlayer) -- Space Block and potentially other mods add inventory
            storage.players[event.player_index].emptyInventory=nil
        end
    end
end)
    
----------------------------------------
script.on_event(defines.events.on_player_created, function(event)
local player = game.players[event.player_index]
if storage.ocfg == nil then
    storage.ocfg = remote.call("oarc_mod", "get_mod_settings")
end
log("on_event::On Player created: " .. player.name)

-- Additions from BraveNewWork OnEvent_on_player_created(event) vf
    if not storage.players then  
        storage.players = {}
    end
    if (#storage.players < event.player_index) then
        log("Zeroing out index :" .. event.player_index .. " drawOnExit to nil for " .. game.players[event.player_index].name)

        storage.players[event.player_index] = {
            crafted = {},
            inventory_items = {},   
            previous_position = player.position,
            drawOnExit = nil,
            characterMode = true,
            inSpawn = true
        }
        storage.spawn[player.index]={x=0,y=0}
    end
    -- Move the player to the game surface immediately.  First time spawning - character has to be deleted
    storage.players[event.player_index].emptyInventory = true    -- postpone until on_player_joined, since space block fills inventory in their on_player_created, which is done AFTER ours.
    if player.character then 
        player.character.destroy()
    end
--    player.teleport({x=0,y=0},  game.surfaces[GAME_SURFACE_NAME])   -- don't use SafeTeleport or double characters show up
    player.set_controller{type=defines.controllers.character,character=getPlayerSurface(player).create_entity{name='character',force=player.force,position=player.position}}

    -- disable light
    player.disable_flashlight()
    -- enable cheat mode - Allows for infinite free crafting which we don't want in normal mode, and done via code when enabled in BNO mode.
    
    -- Set-up a sane default for the quickbar
    if storage.bno.space_block then 
        default_qb_slots[6]="fast-inserter"    
        default_qb_slots[7]="burner-inserter"    
    
        default_qb_slots[10]="landfill"    
        default_qb_slots[17]="iron-chest"  
        default_qb_slots[19]="spaceblock-matter-furnace"  
        
        default_qb_slots[20]="spaceblock-water" 
        default_qb_slots[26]="medium-electric-pole"
        default_qb_slots[27]="small-electric-pole"        
        default_qb_slots[29]="small-lamp"
    end
    for i = 1, 100 do
        if not player.get_quick_bar_slot(i) then
            if default_qb_slots[i] then
                player.set_quick_bar_slot(i, default_qb_slots[i])
            end
        end
    end

    storage.bnw_scenario_version = script.active_mods["brave-new-oarc"]
    -- setup force   vf TODO do this in on_gui_click using new location that comes back from SpawnOptsGuiClick
    -- setupForce(player.force, player.surface, 0, 0, script.active_mods["SeaBlock"])
    if (storage.bno.space_block) then
        TemporaryHelperText(getPlayerSurface(player), "Assemblers produce coal,wood, iron,copper,stone and random items at a slower rate than space matter furnaces", {-22, -10}, TICKS_PER_MINUTE*5, 1.5,{0,.7,1,1})
    end
--    TemporaryHelperText(getPlayerSurface(player), "Check Top left - System Info for more information",                                             {-22, 10}, TICKS_PER_MINUTE*5, 1.5,{0,.7,1,1})
end)


script.on_event(defines.events.on_player_respawned, function(event)
-- log("on_event::on_player_respawned: " .. game.players[event.player_index].name)
end)

script.on_event(defines.events.on_player_left_game, function(event)
log("on_event::on_player_left_game - " .. game.players[event.player_index].name)
    ServerWriteFile("player_events", game.players[event.player_index].name .. " left the game." .. "\n")
    local player = game.players[event.player_index]
    if player.gui.screen.stats_gui then     -- if stats menu is open - close it
        player.gui.screen.stats_gui.destroy() 
    end
    -- If players leave early, say goodbye.
    if (player and (player.online_time < (storage.ocfg.gameplay.minimum_online_time * TICKS_PER_MINUTE))) then
        log("Player left early - removing: " .. player.name)
    else
        if storage.spawn[player.index]~=nil then
            storage.players[player.index].drawOnExit = rendering.draw_text{text=player.name,
                            surface=game.surfaces["nauvis"],
                            target={x=storage.spawn[player.index].x-21, y=storage.spawn[player.index].y+5},
                            color={0.9, 0.7, 0.3, 0.8},
                            scale=20,
                            font="compi",
                            draw_on_ground=true,
                            orientation=0,
                            scale_with_zoom=false,
                            only_in_alt_mode=false}
--            log("Player player.name - drawOnExit created = " .. storage.players[player.index].drawOnExit)
        end
    end
end)

-- script.on_event(defines.events.on_player_removed, function(event)
    -- Player is already deleted when this is called.
-- end)

----------------------------------------
-- On tick events. Stuff that needs to happen at regular intervals.
-- Delayed events, delayed spawns, ...
----------------------------------------
script.on_event(defines.events.on_tick, function(event)
--    FadeoutRenderOnTick()

    if #storage.bno.bno_decon_miners>0 and ((game.tick % (TICKS_PER_SECOND)) == 2) then
        OarcAutoDeconOnTick()
    end

    if storage.bno.warn_biter_attack then
        storage.swarmCheckTick = storage.swarmCheckTick or game.tick
        if (game.tick >= storage.swarmCheckTick) then
            BNOCleanGPSStack()
            storage.swarmCheckTick = game.tick + TICKS_PER_SECOND*2    -- check again in 2 seconds
        end
    end
end)

----------------------------------------
-- Various on "built" events
----------------------------------------
script.on_event(defines.events.on_built_entity, function(event)
--    if storage.bno.enable_anti_grief then
--        SetItemBlueprintTimeToLive(event)
--    end

end)

----------------------------------------
-- Shared chat, so you don't have to type /s
-- But you do lose your player colors across forces.
----------------------------------------
script.on_event(defines.events.on_console_chat, function(event)
    if (event.player_index) then
        ServerWriteFile("server_chat", game.players[event.player_index].name .. ": " .. event.message .. "\n")
    end
--    if (storage.ocfg.enable_shared_chat) then
--        if (event.player_index ~= nil) then
--            ShareChatBetweenForces(game.players[event.player_index], event.message)
--        end
--    end
end)

--script.on_event(defines.events.on_console_command, function(event)
--    if (event.player_index and event.message) then
--        ServerWriteFile("server_console_command", game.players[event.player_index].name .. ": " .. event.message .. "\n")
--    end
--end)


----------------------------------------
-- On Research Finished
-- This is where you can permanently remove researched techs
----------------------------------------
script.on_event(defines.events.on_research_finished, function(event)
    local force = event.research.force

--    for indx,player in pairs(game.connected_players) do
--        if (player.force.name == force.name) then
--            player.print("Your team completed researching [technology=" .. event.research.name .. "]", {r=0, g=102/255, b=0, a=1}) -- green
--            log("Team " .. force.name .. " completed researching [technology=" .. event.research.name .. "]")
--        end
--    end
end)


----------------------------------------
-- On Research Started
----------------------------------------
script.on_event(defines.events.on_research_started, function(event)
    local force = event.research.force
    for indx,player in pairs(game.connected_players) do
        if (player.force.name == force.name) then
            player.print("Your team started researching [technology=" .. event.research.name .. "]", {r=252, g=186, b=3, a=1}) -- Yellow
            log("Team " .. force.name .. " started researching [technology=" .. event.research.name .. "]")
        end
    end
end)


----------------------------------------
-- On unit group finished gathering
-- This is where I remove biter waves on offline players
----------------------------------------
script.on_event(defines.events.on_unit_group_finished_gathering, function(event)
--    if (storage.ocfg.enable_offline_protect) then
--        OarcModifyEnemyGroup(event.group)
--    end
end)

----------------------------------------
-- On player clicked on gps tag
-- Save player's stuff so they don't lose it if they can't get to the corpse fast enough.
----------------------------------------

script.on_event(defines.events.on_player_clicked_gps_tag, function(event)
    local biter, swarmGroup
    if (#storage.swarmGroup > 0) then
        for k,swarm in pairs(storage.swarmGroup ) do
            if ((swarm.startPosition.x == event.position.x) and  (swarm.startPosition.y == event.position.y)) then
                swarmGroup = swarm.group
                break
            end
        end
    end
    if (swarmGroup ~= nil and swarmGroup.valid) then
        for _,member in pairs(swarmGroup.members) do
            if (member.active) then
                biter=member
                break
            end
        end
        if (biter ~=nil) then
            if (biter.valid) then   -- follow the biter swarm with text and camera
--2.0                game.players[event.player_index].zoom_to_world(event.position,  0.5, biter) -- follow a live biter
                local rid1=rendering.draw_text{text=string.format("Swarm coming to kill you, %s.", game.players[event.player_index].name),
                        surface=game.surfaces[GAME_SURFACE_NAME],
                        target=biter,
                        color={1,0.1,0.1,1},
                        scale=3,
                        font="compi",
                        time_to_live=TICKS_PER_SECOND*9,
                        draw_on_ground=false}
                local rid2=rendering.draw_text{text="Press ESC to exit this view.",
                        surface=game.surfaces[GAME_SURFACE_NAME],
                        target=biter,
                        target_offset={0,2},
                        color={1,0.1,0.1,1},
                        scale=2,
                        font="compi",
                        time_to_live=TICKS_PER_SECOND*12,
                        draw_on_ground=false}
--                        table.insert(storage.oarc_renders_fadeout, rid1)            
--                        table.insert(storage.oarc_renders_fadeout, rid2)            
            else    -- show where they were when they formed
--2.0                game.players[event.player_index].zoom_to_world(event.position)                
--                local rid1=rendering.draw_text{text=string.format("Swarm dead, they formed here, %s.", game.players[event.player_index].name),                        surface=game.surfaces[GAME_SURFACE_NAME],                        target={event.position.x, event.position.y},                        color={1,0.1,0.1,0.7},                        scale=3,                        font="compi",                        time_to_live=TICKS_PER_SECOND*9,                        draw_on_ground=false}
--                local rid2=rendering.draw_text{text="Press ESC to exit this view.",                        surface=game.surfaces[GAME_SURFACE_NAME],                        target=biter,{event.position.x, event.position.y+2},                        color={1,0.1,0.1,0.7},                        scale=2,                        font="compi",                        time_to_live=TICKS_PER_SECOND*12,                        draw_on_ground=false}
--                table.insert(storage.oarc_renders_fadeout, rid1)            
--                table.insert(storage.oarc_renders_fadeout, rid2)            
            end
        end
    else    -- show where they were when they formed
--2.0        game.players[event.player_index].zoom_to_world(event.position)
--        local rid1=rendering.draw_text{text=string.format("Swarm dead, they formed here, %s.", game.players[event.player_index].name),                    surface=game.surfaces[GAME_SURFACE_NAME],                    target=event.position,                    color={1,0.1,0.1,0.7},                    scale=3,                    font="compi",                    time_to_live=TICKS_PER_SECOND*9,                    draw_on_ground=false}
--        local rid2=rendering.draw_text{text=string.format("Press ESC to exit this view.", game.players[event.player_index].name),                    surface=game.surfaces[GAME_SURFACE_NAME],                    target={event.position.x, event.position.y+2},                    color={1,0.1,0.1,0.7},                    scale=2,                    font="compi",                    time_to_live=TICKS_PER_SECOND*12,                    draw_on_ground=false}
--         table.insert(storage.oarc_renders_fadeout, rid1)            
--         table.insert(storage.oarc_renders_fadeout, rid2)            
    end
end)

----------------------------------------
-- On Corpse Timed Out
----------------------------------------
script.on_event(defines.events.on_character_corpse_expired, function(event)
    if event.corpse.last_user then
        local player = event.corpse.last_user
        player.print("Your character corpse has expired " .. GetGPStext(player.position))
    end
end)

----------------------------------------
-- On enemies killed
-- For coin generation and stuff
----------------------------------------
script.on_event(defines.events.on_post_entity_died, function(event)
end,
{{filter="type", type = "unit"}, {filter="type", type = "unit-spawner"}, {filter="type", type = "turret"}})


----------------------------------------
-- Scripted auto decon for miners...
----------------------------------------
script.on_event(defines.events.on_resource_depleted, function(event)
    OarcAutoDeconOnResourceDepleted(event)
end)

-----------------------------------------------------------------------------------------------------------
-- Start of Brave New World functions
-----------------------------------------------------------------------------------------------------------

--=================================================================================================
-- @ event: on_player_main_inventory_changed
script.on_event(defines.events.on_player_main_inventory_changed, function(event)
    if storage.creative or storage.players[event.player_index].characterMode then
        return
    end
    local player = game.players[event.player_index]
    if storage.seablocked ==nil then
    else
        if not storage.seablocked then
            -- tiny hack to work around that SeaBlock sets up stuff after BNW on load
            storage.seablocked = true
            -- move everything from the Home rock to the other chest
            local home_rock = getPlayerSurface(player).find_entity("rock-chest", {0.5, 0.5})
            if home_rock then
                for name, count in pairs(home_rock.get_inventory(defines.inventory.chest).get_contents()) do
                    storage.seablock_chest.insert{name = name, count = count}
                end
            end
            home_rock.destroy()
            storage.seablock_chest = nil

            -- and clear the starting items from player inventory
            player.clear_items_inside()
        end
    end
    -- remove any crafted items (and possibly make ghost cursor of item)
    for _, item in pairs(storage.players[event.player_index].crafted) do
        if itemCountAllowed(item.name, item.count, player) == 0 then
            if player.clean_cursor() then
                player.cursor_stack.clear()
            end
        end
--2.0        player.cursor_ghost = game.item_prototypes[item.name]
        player.remove_item(item)
    end
    storage.players[event.player_index].crafted = {}

    -- player is only allowed to carry whitelisted items
    -- everything else goes into entity opened or entity beneath mouse cursor
    -- log("Clearing inventory")
    local inventory_main = player.get_inventory(defines.inventory.god_main)
    local items = {}
    if (inventory_main ~= nil) then	-- vf
        for i = 1, #inventory_main do
            local item_stack = inventory_main[i]
            if item_stack and item_stack.valid_for_read and not item_stack.is_blueprint then
                local name = item_stack.name
                if items[name] then
                    items[name].count = items[name].count + item_stack.count
                else
                    items[name] = {
                        count = item_stack.count,
                        slot = item_stack
                    }
                end
            end
        end
    end
    storage.players[event.player_index].inventory_items = items

    local entity = player.selected or player.opened
    for name, item in pairs(items) do
        local allowed = itemCountAllowed(name, item.count, player)
        local to_remove = item.count - allowed
        if to_remove > 0 then
            dropItems(entity, player, name, to_remove)
            player.remove_item{name = name, count = to_remove}
        end
    end

    checkForStealing(player, entity)
end)

function dropItems(entity, player, name, count)
--    local entity = player.opened or player.selected
    local inserted = 0
    if (storage.enable_oe_debug) then
        log("dropItems- player: ".. player.name .. ", name: " .. name .. " count: " .. count)
    end
    if entity and entity.insert then
        -- in case picking up items from a limited chest, unset limit, insert, then set limit again
        for _, inventory_id in pairs(defines.inventory) do
            local inventory = entity.get_inventory(inventory_id)
            if inventory then
                local barpos = inventory.supports_bar() and inventory.get_bar() or nil
                if inventory.supports_bar() then
                    inventory.set_bar() -- clear bar (the chest size limiter)
                end
                inserted = inserted + inventory.insert{name = name, count = count}
                count = count - inserted
                if inventory.supports_bar() then
                    inventory.set_bar(barpos) -- reset bar
                end
                if count <= 0 then
                    break
                end
            end
        end
        if count > 0 then
            -- try a generic insert (although code above should make this redundant)
            count = count - entity.insert({name = name, count = count})
        end
    end
   if count > 0 then
        -- now we're forced to spill items
        if storage.bno.alien_module and name=="artifact-ore" and entity==nil then
            -- the ore it will just disappear - no one gets it
        else
            entity = entity or storage.forces[player.force.name].roboport    -- dump unknown item to main roboport
        end
        if entity and entity.valid then
            -- This was causing crashes in dropItems if a ghost entity forced a drop, a tank running over stone, or dropping a ghost on stone could do this                  
            if (storage.enable_oe_debug) then
                log("dropItems: Spilling items for: ".. entity.name .. ", type: " .. entity.type .. ", at " .. GetGPStext(entity.position) .. ", entity force: ".. entity.force.name)
            end
            if (entity.name == "oarc-gui") then
                game.players[player.name].print("Close your menu when stealing ore from someone ;)")                    
            elseif (entity.surface == nil) then
                game.players[player.name].print("Send this log to JustGoFly - it shows something that WOULD have crashed, and provides good info to debug what caused it")
                log("dropItems: would have crashed accessing entity.surface - on entity name: " .. entity.name .. " for item: " .. name .. " count: " .. count .. " for player: " .. player.name)
            else
                entity.surface.spill_item_stack{position=entity.position, stack={name = name, count = count}, false, force=entity.force, false}
            end
        end
    end
end

function itemCountAllowed(name, count, player)
    local item = prototypes.item[name]
    local place_type = item.place_result and item.place_result.type
    if name == "red-wire" or name == "green-wire" then
        -- need these for circuitry, one stack is enough
        return math.min(200, count)
    elseif name == "copper-cable" then
        -- need this for manually connecting poles, but don't want player to manually move stuff around so we'll limit it
        return math.min(20, count)
    elseif item.type == "blueprint" or item.type == "deconstruction-item" or item.type == "blueprint-book" or item.type == "selection-tool" or name == "artillery-targeting-remote" or name == "spidertron-remote" or item.type == "upgrade-item" or item.type == "copy-paste-tool" or item.type == "cut-paste-tool" or name == "tl-adjust-capsule" or name == "tl-draw-capsule" or name == "tl-edit-capsule" then
        -- these only place ghosts or are utility items
        return count
    elseif place_type == "car" or place_type == "spider-vehicle" then
        -- let users put down cars & tanks
        return count
    elseif item.place_as_equipment_result then
        -- let user carry equipment
        return count
    elseif string.match(name, ".*module.*") then
        -- allow modules
        return count
    elseif string.match(name, "dt%-cell") then
        -- allow Krastorio 2 special cells
        return count
    elseif string.match(name, ".*fuel%-cell") then
        -- allow Krastorio 2 fuel cells
        return count
    elseif string.match(name, ".*%-fuel") then
        -- allow lexi-aircraft to load fuel cells into aircraft by hand
        return count
    elseif name == "BlueprintAlignment-blueprint-holder" then
        -- temporary holding location for original blueprint, should only ever be one of these.
        return count
        -- allow poles and combinators for circuitissimo
    elseif storage.bno.circuitissimo and (string.match(name, ".*combinator*") or string.match(name, ".*pole")) then
        return math.min(20, count)
    end
    return 0
end

-- use permissions
function preventMining(player)
    configure_permissions(player)
--    if storage.players[player.index].characterMode then 
--        player.force.manual_mining_speed_modifier = 0  -- allow mining
--        EnableTech(player.force, "steel-axe")
--        return
--    else
--    -- prevent mining (this appeared to be reset when loading a 0.16.26 save in 0.16.27)
--        player.force.manual_mining_speed_modifier = -0.99999999 -- allows removing ghosts with right-click
--        if storage.bno.krastorio2 then
--            if storage.players[player.index].characterPlayer then
--                EnableTech(player.force, "kr-iron-pickaxe")
--                EnableTech(player.force, "kr-advanced-pickaxe")
--            else
--                DisableTech(player.force, "kr-iron-pickaxe")
--                DisableTech(player.force, "kr-advanced-pickaxe")
--            end
--        end
--        DisableTech(player.force, "steel-axe") -- researching this upgrades the above manual_mining_speed_modifier by 1 - not good !
--    end
end

--=================================================================================================
-- @ event: on_player_pipette

-- if player hits q over an item, check their inventory to see if they have that item,
-- if they do - load their hand, otherwise put a ghost in their hand
script.on_event(defines.events.on_player_pipette, function(event)
    if storage.creative or storage.players[event.player_index].characterMode then
        return
    end
    local player = game.players[event.player_index]
    local inv = storage.players[event.player_index].inventory_items
    local allowed = 0
    if inv[event.item.name] then
        allowed = itemCountAllowed(event.item.name, inv[event.item.name].count, player)
    end
    if allowed == 0 then
        player.cursor_stack.clear()
        player.cursor_ghost = event.item
    end
end)

--=================================================================================================
-- @ event: on_player_crafted_item
script.on_event(defines.events.on_player_crafted_item, function(event)
    if storage.creative or storage.players[event.player_index].characterMode then
        return
    end
    pcall(function()    -- mostly safeguard seablock from crashing when a player tries to craft a liquid
        game.players[event.player_index].cursor_ghost = event.item_stack.prototype
        event.item_stack.count = 0
    end)
end)

--=================================================================================================
-- @ event: on_player_cursor_stack_changed
script.on_event(defines.events.on_player_cursor_stack_changed, function(event)
    if storage.creative or storage.players[event.player_index].characterMode  then
        return
    end
    local player = game.players[event.player_index]
    local cursor = player.cursor_stack
    if cursor and cursor.valid_for_read then
        local allowed = itemCountAllowed(cursor.name, cursor.count, player)
        local to_remove = cursor.count - allowed
        if to_remove > 0 then
            local entity = player.opened or player.selected
            dropItems(entity, player, cursor.name, to_remove)
            if allowed > 0 then
                cursor.count = allowed
            else
                player.cursor_ghost = cursor.prototype
                player.cursor_stack.clear()
            end
        end
    end

    checkForStealing(player, player.opened)

    -- check if user is in trouble due to insufficient storage
    local alerts = player.get_alerts{type = defines.alert_type.no_storage}
    local out_of_storage = false
    for _, surface in pairs(alerts) do
        for _, alert_type in pairs(surface) do
            for _, alert in pairs(alert_type) do
                local entity = alert.target
                if (entity.name == "construction-robot") then
                    out_of_storage = true
                    local inventory = entity.get_inventory(defines.inventory.robot_cargo)
                    if inventory then
                        for _, item in pairs(inventory.get_contents()) do
                            entity.surface.spill_item_stack{position=entity.position, stack={name = item.name, count = item.count}, force=entity.force}
                        end
                    end
                    entity.clear_items_inside()
                end
            end
        end
    end
    
    if out_of_storage then
        player.print({"out-of-storage", player.name})
        log("Sending out of storage message to: " .. player.name)
    end
end)


script.on_event(defines.events.on_player_fast_transferred, function(event)
    local player = game.players[event.player_index]
    local entity = event.entity
    local itemsStolen
    local items=""

    if (player.force.name ~= entity.force.name) then   -- taking something 
        log("Player: ".. player.name .. " stealing from " .. entity.force.name)
        if not pcall(function()    -- this has been known to crash - on get_contents
            if player.selected then
                itemsStolen = player.selected.get_output_inventory().get_contents()
                for  name, count in pairs(itemsStolen) do
                    items = items ..  name .. " of qty : " ..  count .. " | "
                end
            end

            for idx,p in pairs(game.connected_players) do
    --          if (p.force.name ~= player.force.name) then
                    screamViolationToPlayers(player, entity, items)
    --          end
            end
            log("WTF (on_player_fast_transferred)" .. player.name .. "just took from " .. entity.last_user.name .. " type: [" .. entity.type .. "] " .. entity.name .. " " .. items ..  GetGPStext(entity.position))
        end) then
            log("WTF (on_player_fast_transferred) Attempt to report theft failed")
        end
    end
end)

function screamViolationToPlayers(player, entity, items, logOnly)            
    logOnly = logOnly or false
    local player_who_stole=""

    if not logOnly then
        for idx,p in pairs(game.connected_players) do
            if (p.force.name ~= player.force.name) then
                if p.force.name == "neutral" then
                    log("WTF - " .. entity.last_user.name .. " stealing from neutral, shouldn't be a problem." .. GetGPStext(entity.position))
                else
                    if entity.last_user.name then
                        player_who_stole="just took from " .. entity.last_user.name
                    else -- nil - happens when chest is empty
                        player_who_stole="tried to steal from "
                    end
                    p.print("WTF " .. player.name .. player_who_stole .. " " .. entity.name .. GetGPStext(entity.position))
                    p.play_sound { path = 'wtf' }
                end
            end
        end
    end
    if items then
        log("WTF " .. player.name .. player_who_stole .. " " .. entity.name .. " " .. items ..  GetGPStext(entity.position))
    else
        log("WTF " .. player.name .. player_who_stole .. " " .. entity.name .. " unknown items " ..  GetGPStext(entity.position))
    end
end

-- called when players stack changes
function checkForStealing(player, entity)
    return  -- debug this - test with copy of blueprint, and making a blueprint

--    -- warn players that someone is touching a chest - unless it's a /o player (admin) accessing a chest
--    if (entity) then
--        if (not string.contains(entity.type, "frame")) then     -- menu open
--            if ((not entity.is_blueprint) and not (entity.is_blueprint_book) and entity.is_player()) then    -- admin accessing another players inventory
--                log("WTF (admin function)" .. player.name .. " just took something from inventory of " .. entity.name .. " body! " ..  GetGPStext(entity.position))
--            else
--                -- if the player and we're not processing a menu of ANY MOD 
--                if (player.opened and (pcall(function () return player.opened.gui == nil end))) then    -- menu open ?
--                    -- fast transfer only of accessing a box/entity of another players that is not neutral force
--                    if (pcall(function() return (not player.opened_self and (player.opened.force.name ~= player.force.name) and (player.opened.force.name~="neutral"))end)) then   -- taking something from someone else
--                        log("Debug info: " .. player.name .. " player.opened.name - " .. player.opened.name .. "player force: " .. player.force.name .. ", opened force: " .. player.opened.force.name)
--                        local illegalTypes = {"container", "linked-container", "logistic-container", "furnace", "item", "boiler", "assembling-machine-1", "assembling-machine-2", "assembling-machine-3", "car"}
--                        local reported=false
--                        for index, value in ipairs(illegalTypes) do
--                            if value == entity.type then
--                                reportViolationToPlayers(player, entity)
--                                reported=true
--                            end
--                        end
--                        if not reported then
--                            reportViolationToPlayers(player, entity, true)
--                        end
--                    end
--                end
--            end
--        end
--    end
end

function reportViolationToPlayers(player, entity, logOnly)
    local items = ""
    if entity.is_entity_with_owner then
        logOnly = logOnly or false
        local stolenInv = entity.get_output_inventory()
        itemsStolen = {}
        for i=1,#stolenInv do
            if stolenInv[i] and stolenInv[i].valid and stolenInv[i].valid_for_read then
                if itemsStolen[stolenInv[i].name] == nil then
                    itemsStolen[stolenInv[i].name] = stolenInv[i].count
                else
                    itemsStolen[stolenInv[i].name] = itemsStolen[stolenInv[i].name] + stolenInv[i].count
                end
            end
        end
        for  name, count in pairs(itemsStolen) do
            items = items ..  name .. " qty remaining : " ..  count .. " | "
        end
        if not logOnly then
            for idx,p in pairs(game.connected_players) do
                if (p.force.name ~= player.force.name) then
                    p.print("WTF " .. player.name .. "just accessed " .. entity.last_user.name .. " " .. entity.name .. " " .. items ..  GetGPStext(entity.position))
                    p.play_sound { path = 'wtf' }
                end
            end
        end
        log("WTF (checkForStealing)" .. player.name .. "just accessed " .. entity.last_user.name .. " type: " .. entity.type .. " <" .. entity.name .. "> " .. items ..  GetGPStext(entity.position))
    end
end


--=================================================================================================
-- @ event: on_entity_died
script.on_event(defines.events.on_entity_died, function(event)
    local playerThatDied=nil
    if storage.creative then
        return
    end
    local entity = event.entity
    -- check if roboport was destroyed - the backer_name is the friendly name given to each roboport - the main one is named after to player
    if (entity.name=="roboport-bno") then
        log("BNO Roboport died for " .. entity.force.name .. " at " .. GetGPStext(entity.position))
        for name,player in pairs(game.players) do
            local SP=entity.position
--log("Player index:" .. tostring(player.index))
--log ("storage.spawn:" .. serpent.block(storage.spawn))
--log ("entity.position:" .. serpent.block(SP))
            -- it has to be at this location - could also compare (entity.backer_name == player.name) 
            if (GetGPStext(storage.spawn[player.index]) == GetGPStext(SP)) then 
                playerThatDied=storage.spawn[player.index]   -- capture the player that died
                SendBroadcastMsg("Oh No " .. player.name .. " on '" .. entity.force.name ..  "' Gone like a fart in the wind")
                log("Force DIED: " .. entity.force.name .. ", player: " .. player.name )
                log ("player: " .. player.name .. " at " ..GetGPStext(storage.spawn[player.index]) .. " Died due to the starting roboport being destroyed.")
                log ("and entity at " .. GetGPStext(SP))
                SendBroadcastMsg("Our buddy " .. player.name .. " on force: '" .. entity.force.name .. "' Died due to the starting roboport being destroyed.")        
                SendMsg(player.name, "Sorry '" .. player.name .. "' you LOSE! Rejoin if you like, and give it another try")
                storage.spawn[player.index]=nil  -- solves a race condition - player dies, then roboport dies, on respawn - crash without this in SafeTeleport
                remote.call("oarc_mod", "remove_or_reset_player", player, not player.connected)
            end
        end
        -- other bno roboports will die - no need to kill off the player unless it's the main one.
        if playerThatDied ~= nil then    -- this crashed once when it was nil - so fix here
            for name, player in pairs(game.connected_players) do
                if (player.index == playerThatDied.index) then
                    player.play_sound { path = 'you-lost' }  -- if the player that died is still online - play random sound
                else
                    player.play_sound { path = 'player-lost' }
                end
            end
        end
    end
    if entity.name=="character" then
        game.set_game_state{game_finished = false, player_won = false, can_continue = true, }
    end
end)

--=================================================================================================
-- @ event: on_player_driving_changed_state
DISTANCE_BETWEEN_BUDDIES = 10

script.on_event(defines.events.on_player_driving_changed_state, function(event)
    local player=game.players[event.player_index]
    log("Changed driver state: " .. event.entity.name)
    -- distance between two points √((x2 – x1)² + (y2 – y1)²)
    local distance = getDistance(event.entity.position, storage.players[event.player_index].previous_position)
    if (distance > DISTANCE_BETWEEN_BUDDIES*32 * 5) then
        log(string.format("msg: %s, sent to %d players", msg, #game.connected_players-1))
        for idx,p in pairs(game.connected_players) do
            if (p.index ~= player.index) then
                p.print("WTF is this guy (" .. player.name .. ") doing teleporting " .. distance .. " tiles " ..  GetGPStext(event.entity.position))
                p.play_sound { path = 'wtf' }
            end
        end        
        log("WTF Player : " .. player.name .. " teleported " .. distance .. " tiles and entered a vehicle far from his main base to x: " .. event.entity.position.x .. ", y: " .. event.entity.position.y)
    end
end)

--=================================================================================================
-- @ event: on_player_changed_position
script.on_event(defines.events.on_player_changed_position, function(event)
    if storage.creative then
        return
    end
	-- We might be waiting for base to render vf
    if not storage.forces then
		return
	end
    -- some mods teleport a player to a new surface, new position, allow this to happen
    if (game.players[event.player_index].surface.name == HOLDING_PEN_SURFACE_NAME) then
        return
    end
    -- enable moving around at 0,0
    if (event.player_index > #storage.players) then
        return
    end
    local player = game.players[event.player_index]
    -- TODO: really shouldn't have to do this so often (can we do it in migrate function?)
    -- preventMining(player)

    local config = storage.forces[player.force.name]
    local x_chunk = math.floor(player.position.x / CHUNK_SIZE)      -- 32
    local y_chunk = math.floor(player.position.y / CHUNK_SIZE)      -- 32
    -- prevent player from exploring, unless in a vehicle
    if storage.spawning then
        log("Player spawning - override the safeguard - allow spawning here")
        storage.spawning = false
    else
        if not player.vehicle then
            local charted = function(x, y)
            local surface = getPlayerSurface(player)
               return player.force.is_chunk_charted(surface, {x, y}) and
                  (player.force.is_chunk_charted(surface, {x - 2, y - 2}) or not surface.is_chunk_generated({x - 2, y - 2})) and
                  (player.force.is_chunk_charted(surface, {x - 2, y + 2}) or not surface.is_chunk_generated({x - 2, y + 2})) and
                  (player.force.is_chunk_charted(surface, {x + 2, y - 2}) or not surface.is_chunk_generated({x + 2, y - 2})) and
                  (player.force.is_chunk_charted(surface, {x + 2, y + 2}) or not surface.is_chunk_generated({x + 2, y + 2}))
            end
--            if player.render_mode == defines.render_mode.game then      -- player not in map view ?
                if not player.character then 
                    if not charted(math.floor(player.position.x / CHUNK_SIZE), math.floor(player.position.y / CHUNK_SIZE)) then -- 32  32
                        -- can't move here, chunk not charted
                        local prev_pos = storage.players[event.player_index].previous_position

                        if charted(math.floor(player.position.x / CHUNK_SIZE), math.floor(prev_pos.y / CHUNK_SIZE)) then    -- 32 32
                            -- we can move here, though
                            prev_pos.x = player.position.x
                        elseif charted(math.floor(prev_pos.x / CHUNK_SIZE), math.floor(player.position.y / CHUNK_SIZE)) then    -- 32 32
                            -- or here
                            prev_pos.y = player.position.y
                        end
                        -- teleport player to (possibly modified) prev_pos
                        if (not player.character) then
                            player.teleport(prev_pos)
                            player.print("Fog of War limit! Drop a radar to extend your reach", {sound=defines.print_sound.never})
                            player.play_sound { path = 'Nu-uhh' }
                        end
                    end
                end
--            end
        end
    end
    -- save new player position
    storage.players[event.player_index].previous_position = player.position
end)


--script.on_event(defines.events.on_runtime_mod_setting_changed, function(event)
--    log("events.on_runtime_mod_setting_changed: setting " .. event.setting .. " for: " .. game.players[event.player_index].name .. "type: " .. event.setting_type)
--    log("Physical setting changed to: " .. settings.get_player_settings(event.player_index)["bno-share-chart"])
--end)

--=================================================================================================
-- @ function: change_bots
function change_bots()
local myConBot
local myLogiBot
	-- body
    myConBot = table.deepcopy(data.raw["construction-robot"]["construction-robot"])
    myConBot.speed = 0.5
    myConBot.minable = {mining_time = 10, result = "construction-robot"}
    myConBot.max_energy = "4MJ"
    myConBot.energy_per_tick = "0.0005kJ"
    myConBot.energy_per_move = "0.2kJ"
    myConBot.destructible = false
    data:extend({myConBot})

    myLogiBot = table.deepcopy(data.raw["logistic-robot"]["logistic-robot"])
    myLogiBot.minable = {mining_time = 10, result = "logistic-robot"}
    myLogiBot.max_energy = "4MJ"
    myLogiBot.energy_per_tick = "0.0005kJ"
    myLogiBot.energy_per_move = "0.2kJ"
    myLogiBot.speed = 0.5
    myLogiBot.destructible = false
    data:extend({myLogiBot})
end

--=================================================================================================
-- @ event: CustomInputEvent
-- Docs: https://lua-api.factorio.com/latest/events.html#CustomInputEvent
-- Taken directly from remove-biter-corpses from RedRafe mod
script.on_event("remove-corpses", function(event)
  local blacklist = {
      'character-corpse',
      'transport-caution-corpse'  -- Transport Drones
      }
      local player = game.players[event.player_index]
      if not player then return end
      local radius = 32
      -- settings.get_player_settings(event.player_index)['rbc-radius'].value
      local surface = SafeSurface(player)

      local bodyCount = 0
      for ___, entity in pairs(surface.find_entities_filtered{
        position = player.position,
        radius = radius,
        type = 'corpse'}
      ) do 
     if not tableContains(blacklist, entity.name) then 
         entity.destroy()
         bodyCount = bodyCount + 1
     end
  end
  if bodyCount > 0 then 
    player.print('Removed ' .. tostring(bodyCount) .. ' corpses', {r=254/255, g=255/255, b=10/255, a=1})
  else 
    player.print('Already Clear', {r=49/255, g=190/255, b=48/255, a=1})
  end
end)
  
--=================================================================================================
-- Remove Alt-Scroll ability
script.on_event("suppress-alt-scroll", function(event)
  -- Do nothing or add custom behavior here
end)

script.on_event("suppress-alt-scroll-down", function(event)
  -- Also do nothing or custom behavior for scroll down
end)
--=================================================================================================
script.on_configuration_changed(function(event)
    local player = game.players[event.player_index]
    local old_ctlr = event.old_type
    local cur_ctrlr = player.controller_type
end)

--=================================================================================================
script.on_event("oarc-mod-on-player-spawned", function(event)
    local player = game.players[event.player_index]

    local first_spawn=event.first_spawn
    local is_host = event.is_host
    log("playerSpawned: " .. player.name .. ", force: " .. player.force.name)

   if storage.bno.space_block then
        for _,v in ipairs(SPACE_BLOCK_RECIPES_REMOVE_ALL_PLAYERS) do
            RemoveRecipe(player.force, v.r)
        end           

        -- remove more items from research if you are Space Block and BNO Player
        if storage.players[player.index].characterMode then
            for _,v in ipairs(SPACE_BLOCK_UNLOCKED_TECHNOLOGIES_CHAR) do
                EnableTech(player.force, v.t)
            end
        else
            for _,v in ipairs(SPACE_BLOCK_LOCKED_TECHNOLOGIES_BNO) do
                DisableTech(player.force, v.t)
            end
        end
    end

    if storage.bno.stranded_block_2 then
        for _,v in ipairs(STRANDED_RECIPES_REMOVE_ALL_PLAYERS) do
            RemoveRecipe(player.force, v.r)
        end
    end
    -- Render some welcoming text...
    -- DisplayWelcomeGroundTextAtSpawn(player, delayedSpawn.pos)	

    -- look into using the unit_number and look at  remote.call("oarc_mod", "get_player_home_spawn")
    local pos = player.position
    -- move base above player by 10 tiles, even if BNO player, a character player might join their base and spawn inside the Roboport
    pos.y=pos.y-10
    log("player name: " .. player.name .. " spawning at " .. GetGPStext(pos))
    storage.spawn[player.index] = pos  -- save the starting position, this is how we determine who died when a starting roboport is killed

    GenerateExtraStartingResources(game.surfaces["nauvis"], pos)

    -- Render Brave New World Items - vf
    if (is_host) then
        log(player.name .. " is host of this team, so building base")
        setupBNWForce(player, pos.x, pos.y)
    else
        log(player.name .. " is NOT host of this team")
    end
    -- Chart the area.
--    ChartArea(player.force, pos, math.ceil(storage.ocfg.spawn_config.gen_settings.land_area_tiles/CHUNK_SIZE), player.character.surface)
    SetupCharacterOrBNOPlayer(player)
end)


--=================================================================================================
script.on_event("oarc-mod-on-player-reset", function(event)
    local player = game.players[event.player_index]
    storage.custom_tabs_table[player.index].firstRocketOT=nil   -- clear first launch variable
    empty_players_inventory(player)
    if player.character == nil then
        if not pcall(function() return (player.create_character()) end) then
            log("oarc-mod-on-player-reset unable to create_character - protected from crash")
        end
    end
    player.cheat_mode=false
end)


--=================================================================================================
script.on_configuration_changed(function(event)
    local new = script.active_mods["brave-new-oarc"]
    if storage.custom_tabs_table == nil then
        storage.custom_tabs_table = {}
    end
    if new ~= nil then
        local old = storage.scenario_version
        if old ~= new then
            game.reload_script()
            storage.scenario_version = new
--            storage.custom_tabs_table[player.index].firstRocketOT
        end
    end
end)