-- Thanks to Bits-Orio for getting me started on Stats table!
require("bno_utils")

BNO_STATS_TAB_NAME = "Stats"
BNO_SETTINGS_TAB_NAME = "BNO Settings"

function format_number_commas(num)
  return tostring(num):reverse():gsub("%d%d%d", "%1,"):reverse():gsub("^,", "")
end

local function format_number(num)
    num = math.ceil(num)
    if num >= 1000000 then
        return string.format("%dM", num / 1000000)
    elseif num >= 1000 then
        return string.format("%dk", num / 1000)
    else
        return tostring(num)
    end
end

function create_dropdown(parent, items, index)
    local dropdown = parent.add{ name="stats_dropdown", type = "drop-down" }
    for indx,item in pairs(items) do
        dropdown.add_item(item.richTextItem)
    end
    dropdown.selected_index = index
    return dropdown
end

-- columnStatItems - contains sets of 7 items to put across the dialog of items and their rich text string
-- statItems is an array that populates the dropdown names of the groups of 7 items
-- stats_table is populated with the stats for each of the 7 items to display based on itemIndx
-- itemIndx is the index of the drop down that chooses the array of items to display
-- 

local numItemColumns=9
function create_stats_gui(player, statItems, columnStatItems, stats_table, itemIndx)
-- dialog based code
--    if player.gui.screen.stats_gui then
--        player.gui.screen.stats_gui.destroy()
--    end
--    local dialog = player.gui.screen.add{
--        type = "frame",
--        name = "stats_gui",
--        caption = "Production Statistics",
--        direction = "vertical"
--    }
    local dialog=storage.custom_tabs_table[player.index].stats_content
    dialog.clear()
     --dialog.auto_center = true

    local scroll_pane = dialog.add{type = "scroll-pane", vertical_scroll_policy = "auto",horizontal_scroll_policy = "auto"}
    scroll_pane.style.maximal_height = 800
    local table = scroll_pane.add{ type = "table", column_count = numItemColumns+1 }
    -----------------------------------------------------------------------------------------------
    --= = = = = = = = = = = = DROP LIST BUTTON and room for 7 more buttons  = = = = = = = = = = = = 
    -- buttons / drop lists ...
    local buttons = dialog.add{type = "table", column_count = numItemColumns+1 }
    -- build drop list from table
    for indx, items in pairs (statItems) do
        if items.item == item then 
            index = indx
            break
        end
    end
    create_dropdown(table, statItems, itemIndx)
    -- = = = = = = = 
    -- 10 boxes here
    -- space 2
    for i=1, 2 do
        local spacer = table.add{type = "label", caption = ""}
        spacer.style.minimal_width = 120
    end
    -- game time 1
    table.add{type = "label", caption = "Game Length: " .. return_time(game.ticks_played)}
    -- space 7
    for i=1, numItemColumns-3 do
        local spacer = table.add{type = "label", caption = ""}
        spacer.style.minimal_width = 120
    end
    -- = = = = = = = 
    -----------------------------------------------------------------------------------------------
    --= = = = = = = = = = = = Statistics header - Force Name and 7 data items  = = = = = = = = = = = = 
    -- Force header
    local label = table.add{type = "label", caption = "[img=entity/character-corpse] Force"}
    label.style.font = "default-bold"
    label.style.font_color = {r = 0, g = 1, b = 0}
    label.style.minimal_width = 120
    -- each data type header
    for i=1, numItemColumns do
        local label = table.add{type = "label", caption = columnStatItems[itemIndx][i].richTextItem}
        label.style.font = "default-bold"
        label.style.font_color = {r = 0, g = 1, b = 0}
        label.style.minimal_width = 120
        label.style.horizontal_align = "right"
    end
    --= = = = = = = = = = = = = = = = = Statistics Data  = = = = = = = = = = = = = = = = = = = = = = 
    -- drop list data
    for _, stat in ipairs(stats_table) do
        local count_label
        if stat.force then
            count_label = table.add{type = "label", caption = stat.force}
        else
            count_label = table.add{type = "label", caption = stat.data}
            count_label.style.horizontal_align = "right"
        end
        count_label.style.minimal_width = 120
    end
    -----------------------------------------------------------------------------------------------
   
    -- close button
-- dialog based code
--    local button = table.add{ type= "button", name="stats_close_stats_gui", caption = "Close"}
--     button.style.minimal_width = 120
--    dialog.force_auto_center()
end

-- itemIndx 1..2 index's into 2x7 array of items
function buildStatsTable(player, itemIndx, forDisplay)
    local statItems = {
        {item="Resources"                    ,richTextItem="[item=iron-plate] Resources", type = "stats"},
        {item="Roboport"                     ,richTextItem="[item=construction-robot] Circuits & Bot items", type = "stats"},
        {item="Production"                   ,richTextItem="[item=oil-refinery] Production Machines", type = "stats"},
        {item="Science"                      ,richTextItem="[item=automation-science-pack] Science", type = "stats"},
        {item="Power"                        ,richTextItem="[item=nuclear-reactor] Power", type = "stats"},
        {item="Time"                         ,richTextItem="[img=quantity-time] Time", type = "stats"},
    }
    local columnStatItems = {
        {
            {item="iron-plate"               ,richTextItem="[item=iron-plate] Iron Plates", type = "stats"},
            {item="copper-plate"             ,richTextItem="[item=copper-plate] Copper Plates", type = "stats"},
            {item="steel-plate"              ,richTextItem="[item=steel-plate] Steel Plates", type = "stats"},
            {item="landfill"                 ,richTextItem="[item=landfill] Landfill", type = "stats"},
            {item="plastic-bar"              ,richTextItem="[item=plastic-bar] Plastic Bar", type = "stats"},
            {item="sulfur"                   ,richTextItem="[item=sulfur] Sulfur", type = "stats"},
            {item="battery"                  ,richTextItem="[item=battery] Battery", type = "stats"},
            {item="crude-oil"                ,richTextItem="[fluid=crude-oil] Crude Oil", type = "fluid"},
            {item=nil, richTextItem=nil, type = nil},
        },
        {
            {item="electronic-circuit"       ,richTextItem="[item=electronic-circuit] Green Circuits", type = "stats"},
            {item="advanced-circuit"         ,richTextItem="[item=advanced-circuit] Red Circuits", type = "stats"},
            {item="processing-unit"          ,richTextItem="[item=processing-unit] Blue Circuits", type = "stats"},
            {item="engine-unit"              ,richTextItem="[item=engine-unit] Yellow Engines", type = "stats"},
            {item="electric-engine-unit"     ,richTextItem="[item=electric-engine-unit] Red Engines", type = "stats"},
            {item="low-density-structure"    ,richTextItem="[item=low-density-structure] Low Density Structure", type = "stats"},
            {item="flying-robot-frame"       ,richTextItem="[item=flying-robot-frame] Flying Robot Frame", type = "stats"},
            {item="construction-robot"       ,richTextItem="[item=construction-robot] Construction Robots", type = "stats"},
            {item="logistic-robot"           ,richTextItem="[item=logistic-robot] Logistic Robots", type = "stats"},
        },
        {
            {item="oil-refinery"             ,richTextItem="[item=oil-refinery] Oil Refineries", type = "stats"},
            {item="chemical-plant"           ,richTextItem="[item=chemical-plant] Chemical Plants", type = "stats"},
            {item="centrifuge"               ,richTextItem="[item=centrifuge] Centrifuges", type = "stats"},
            {item="lab"                      ,richTextItem="[item=lab] Labs", type = "stats"},
            {item="beacon"                   ,richTextItem="[item=beacon] Beacons", type = "stats"},
            {item="nuclear-reactor"          ,richTextItem="[item=nuclear-reactor] Nuclear Reactors", type = "stats"},
            {item="rocket-silo"              ,richTextItem="[item=rocket-silo] Rocket Silos", type = "stats"},
            {item="total assemblers 1..3"    ,richTextItem="[item=assembling-machine-1] assembling machines 1..3", type = "stats"},
            {item="electric mining drills"   ,richTextItem="[item=electric-mining-drill] electric-mining-drill", type = "stats"},
        },
        {
            {item="automation-science-pack"  ,richTextItem="[item=automation-science-pack] Automation", type = "stats"},
            {item="logistic-science-pack"    ,richTextItem="[item=logistic-science-pack] Logistics", type = "stats"},
            {item="military-science-pack"    ,richTextItem="[item=military-science-pack] Military", type = "stats"},
            {item="chemical-science-pack"    ,richTextItem="[item=chemical-science-pack] Chemical", type = "stats"},
            {item="production-science-pack"  ,richTextItem="[item=production-science-pack] Production", type = "stats"},
            {item="utility-science-pack"     ,richTextItem="[item=utility-science-pack] Utility", type = "stats"},
            {item="space-science-pack"       ,richTextItem="[item=space-science-pack] Space", type = "stats"},
            {item=nil, richTextItem=nil, type = nil},
            {item=nil, richTextItem=nil, type = nil},
        },
        {
            {item="uranium-ore"              ,richTextItem="[item=uranium-ore] Uranium Ore", type = "stats"},
            {item="uranium-238"              ,richTextItem="[item=uranium-238] Uranium 238", type = "stats"},
            {item="uranium-235"              ,richTextItem="[item=uranium-235] Uranium 235", type = "stats"},
            {item="depleted-uranium-fuel-cell" ,richTextItem="[item=depleted-uranium-fuel-cell] Depleted Uranium", type = "stats"},
            {item="rocket-fuel"              ,richTextItem="[item=rocket-fuel] Rocket Fuel",             type = "stats"},
            {item="nuclear-fuel"             ,richTextItem="[item=nuclear-fuel] Nuclear Fuel",           type = "stats"},
            {item="uranium-fuel-cell"        ,richTextItem="[item=uranium-fuel-cell] Uranium Fuel Cell", type = "stats"},
            {item=nil, richTextItem=nil, type = nil},
            {item=nil, richTextItem=nil, type = nil},
        },
        {
            {item="num_players"              ,richTextItem="Num Players",   type = "num players"},
            {item="online_time"              ,richTextItem="Online Time",   type = "ticks"},
            {item="last_online"              ,richTextItem="Last Online",   type = "ticks"},
            {item="afk_time"                 ,richTextItem="AFK Time",      type = "ticks"},
            {item="First Launch"             ,richTextItem="First [space-age] Launch", type = "stats"},
            {item="Rockets Launched"        , richTextItem="[item=satellite] Launched", type = "stats"},
            {item=nil, richTextItem=nil, type = nil},
            {item=nil, richTextItem=nil, type = nil},
            {item=nil, richTextItem=nil, type = nil},
        }
    }
    if storage.bno.krastorio2 then
        local addToStat = table.insert(statItems, 5, {item="Krastorio Science",richTextItem="[img=virtual-signal/kr-krastorio] Krastorio Science", type = "stats"})
        local addToStat = table.insert(columnStatItems,5,         {
            {item="kr-blank-tech-card"       ,richTextItem="[item=kr-blank-tech-card] blank tech",   type = "stats"},
            {item="kr-basic-tech-card"       ,richTextItem="[item=kr-basic-tech-card] basic tech",   type = "stats"},
            {item="kr-matter-tech-card"      ,richTextItem="[item=kr-matter-tech-card] matter tech", type = "stats"},
            {item="kr-advanced-tech-card"    ,richTextItem="[item=kr-advanced-tech-card] advanced tech", type = "stats"},
            {item="kr-singularity-tech-card" ,richTextItem="[item=kr-singularity-tech-card] singularity tech", type = "stats"},
            {item=nil, richTextItem=nil, type = nil},
            {item=nil, richTextItem=nil, type = nil},
            {item=nil, richTextItem=nil, type = nil},
            {item=nil, richTextItem=nil, type = nil},
        })
    end
    local stats_table = {}
--    local quality= {{item="normal"}, {item="uncommon"}, {item="rare"}, {item="epic"}, {item="legendary"} }
    log("Stats selection by <" .. player.name .. "> : " .. itemIndx)

    for _, force in pairs(game.forces) do
--       local ignoredalienmodulefactions = { enemy=true, neutral=true, _ABANDONED_=true, _DESTROYED_=true, player=true} 
--       if not ignoredalienmodulefactions[force.name] then
         if (#force.players > 0) then   -- all above ignore forces will have zero players - will also eliminate Main Force if needed
            table.insert(stats_table, {force=force.name})
            local stats = force.get_item_production_statistics(getPlayerSurface(player))
            local fluids = force.get_fluid_production_statistics(getPlayerSurface(player))
            
            for i=1, numItemColumns do
                local item = columnStatItems[itemIndx][i].item
                if item then
                    if columnStatItems[itemIndx][i].type == "stats" then
                        local all_time=0
                        if item == "total assemblers 1..3" then
                            a1 = stats.input_counts["assembling-machine-1"] or 0
                            a2 = stats.input_counts["assembling-machine-2"] or 0
                            a3 = stats.input_counts["assembling-machine-3"] or 0
                            all_time = a1 + a2 + a3
                        else
                            all_time = stats.input_counts[item] or 0
                        end
                        table.insert(stats_table, {data=format_number_commas(all_time)})
                    elseif columnStatItems[itemIndx][i].type == "fluid" then
                        all_time = fluids.input_counts[item] or 0
                        table.insert(stats_table, {data=format_number_commas(all_time)})
                    elseif columnStatItems[itemIndx][i].type == "assemblers" then
                        if i<4 then
                            local all_time = stats.input_counts[item] or 0
                            table.insert(stats_table, {data=format_number_commas(all_time)})
                        else
                            if storage.bno.brave_new_assembling_machines then
                                if i<7 then -- 6 items
                                    local all_time = stats.input_counts[item] or 0
                                    table.insert(stats_table, {data=format_number_commas(all_time)})
                                else
                                    table.insert(stats_table, {data = ""})
                                end
                            else
                                table.insert(stats_table, {data = ""})
                            end
                        end
                    elseif columnStatItems[itemIndx][i].type == "num players" then
                    -- just do it all at once
                        local online_time=0
                        local afk_time=0
                        local last_online = 0
                        local num_players = #force.players
                        local first_rocket_time = 0
                        local num_rockets_launched = 0
                        table.insert(stats_table, {data = format_number_commas(num_players)})
                        for _,player in pairs(force.players) do
                            online_time = online_time + player.online_time  -- combined team online time
                            afk_time = afk_time + player.afk_time           -- combined afk time
                            if player.last_online > last_online then        -- last person online time
                                last_online = player.last_online
                            end
                            -- first rocket ?
                            if first_rocket_time== 0 then
                                if storage.custom_tabs_table[player.index].firstRocketOT == nil then
                                    storage.custom_tabs_table[player.index].firstRocketOT = 0
                                end
                                first_rocket_time = storage.custom_tabs_table[player.index].firstRocketOT
                            else
                                if first_rocket_time > storage.custom_tabs_table[player.index].firstRocketOT then
                                    first_rocket_time = storage.custom_tabs_table[player.index].firstRocketOT
                                end
                            end
                        end
                        table.insert(stats_table, {data = return_time(online_time)})
                        table.insert(stats_table, {data = return_time(last_online)})
                        table.insert(stats_table, {data = return_time(afk_time)})
                        if first_rocket_time and first_rocket_time>0 then
                            table.insert(stats_table, {data = return_time(first_rocket_time)})
                        else    
                            table.insert(stats_table, {data = ""})
                        end
--                        num_rocket_launched = force.items_launched[
                        
                        local numSatellitesLaunched = force.items_launched["satellite"]
                        if numSatellitesLaunched == nil then 
                            table.insert(stats_table, {data = ""})
                        else
                            table.insert(stats_table, {data = tostring(numSatellitesLaunched)})
                        end
                        table.insert(stats_table, {data = ""})
                        table.insert(stats_table, {data = ""})
                        table.insert(stats_table, {data = ""})
                        break -- what a cheater I am - lol !
                    end
                else
                    table.insert(stats_table, {data = ""})
                end
            end
        end
    end
--    table.sort(stats_table, function(a, b) return a.all_time > b.all_time end)
    create_stats_gui(player, statItems, columnStatItems, stats_table, itemIndx)
end

function firstRocketLaunched(player, tick)
    -- storage.custom_tabs_table[player.index].firstRocketAt = tick
    storage.custom_tabs_table[player.index].firstRocketOT = player.online_time
end